"""
HappyAccidents API Wrapper

:license: Apache License 2.0, see LICENSE for more details.
"""

__title__ = "happyaccidentsapi"
__author__ = "hoopengo"
__license__ = "Apache License 2.0"
__version__ = "0.0.0"

from ._happyaccidentsapi import *
from .enums import *
from .models import *
